# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib, urlparse, parse_qsl, assertRaisesRegex
from libcloud.loadbalancer.base import Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import Provider
from libcloud.test.file_fixtures import LoadBalancerFileFixtures
from libcloud.loadbalancer.providers import get_driver
from libcloud.loadbalancer.drivers.cloudstack import CloudStackLBDriver

try:
    import simplejson as json
except ImportError:
    import json


class CloudStackLBTests(unittest.TestCase):
    def setUp(self):
        CloudStackLBDriver.connectionCls.conn_class = CloudStackMockHttp

        CloudStackLBDriver.path = "/test/path"
        CloudStackLBDriver.type = -1
        CloudStackLBDriver.name = "CloudStack"
        self.driver = CloudStackLBDriver("apikey", "secret")
        CloudStackMockHttp.fixture_tag = "default"
        self.driver.connection.poll_interval = 0.0

    def test_user_must_provide_host_and_path(self):
        CloudStackLBDriver.path = None
        CloudStackLBDriver.type = Provider.CLOUDSTACK

        expected_msg = (
            "When instantiating CloudStack driver directly "
            + "you also need to provide host and path argument"
        )
        cls = get_driver(Provider.CLOUDSTACK)

        assertRaisesRegex(self, Exception, expected_msg, cls, "key", "secret")

        try:
            cls("key", "secret", True, "localhost", "/path")
        except Exception:
            self.fail("host and path provided but driver raised an exception")

    def test_list_supported_algorithms(self):
        algorithms = self.driver.list_supported_algorithms()

        self.assertTrue(Algorithm.ROUND_ROBIN in algorithms)
        self.assertTrue(Algorithm.LEAST_CONNECTIONS in algorithms)

    def test_list_balancers(self):
        balancers = self.driver.list_balancers()
        for balancer in balancers:
            self.assertTrue(isinstance(balancer, LoadBalancer))

    def test_create_balancer(self):
        members = [Member(1, "1.1.1.1", 80), Member(2, "1.1.1.2", 80)]
        balancer = self.driver.create_balancer(
            name="test", algorithm=Algorithm.ROUND_ROBIN, members=members
        )
        self.assertTrue(isinstance(balancer, LoadBalancer))

    def test_destroy_balancer(self):
        balancer = self.driver.list_balancers()[0]
        self.driver.destroy_balancer(balancer)

    def test_balancer_attach_member(self):
        balancer = self.driver.list_balancers()[0]
        member = Member(id=1234, ip="1.1.1.1", port=80)
        balancer.attach_member(member)

    def test_balancer_detach_member(self):
        balancer = self.driver.list_balancers()[0]
        member = balancer.list_members()[0]
        balancer.detach_member(member)

    def test_balancer_list_members(self):
        balancer = self.driver.list_balancers()[0]
        members = balancer.list_members()
        for member in members:
            self.assertTrue(isinstance(member, Member))
            self.assertEqual(member.balancer, balancer)


class CloudStackMockHttp(MockHttp, unittest.TestCase):
    fixtures = LoadBalancerFileFixtures("cloudstack")
    fixture_tag = "default"

    def _load_fixture(self, fixture):
        body = self.fixtures.load(fixture)
        return body, json.loads(body)

    def _test_path(self, method, url, body, headers):
        url = urlparse.urlparse(url)
        query = dict(parse_qsl(url.query))

        self.assertTrue("apiKey" in query)
        self.assertTrue("command" in query)
        self.assertTrue("response" in query)
        self.assertTrue("signature" in query)

        self.assertTrue(query["response"] == "json")

        del query["apiKey"]
        del query["response"]
        del query["signature"]
        command = query.pop("command")

        if hasattr(self, "_cmd_" + command):
            return getattr(self, "_cmd_" + command)(**query)
        else:
            fixture = command + "_" + self.fixture_tag + ".json"
            body, obj = self._load_fixture(fixture)
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _cmd_queryAsyncJobResult(self, jobid):
        fixture = "queryAsyncJobResult" + "_" + str(jobid) + ".json"
        body, obj = self._load_fixture(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import copy
import json
import unittest
import email.utils
from io import BytesIO
from unittest import mock
from unittest.mock import Mock, PropertyMock

import pytest

from libcloud.test import StorageMockHttp
from libcloud.utils.py3 import StringIO, httplib
from libcloud.common.types import InvalidCredsError
from libcloud.storage.base import Object, Container
from libcloud.test.secrets import STORAGE_GOOGLE_STORAGE_PARAMS
from libcloud.common.google import GoogleAuthType
from libcloud.storage.drivers import google_storage
from libcloud.test.file_fixtures import StorageFileFixtures
from libcloud.test.storage.test_s3 import S3Tests, S3MockHttp
from libcloud.test.common.test_google import GoogleTestCase

CONN_CLS = google_storage.GoogleStorageConnection
JSON_CONN_CLS = google_storage.GoogleStorageJSONConnection

TODAY = email.utils.formatdate(usegmt=True)


def _error_helper(code, headers):
    message = httplib.responses[code]
    body = {
        "error": {"errors": [{"code": code, "message": message, "reason": message}]},
    }
    return code, json.dumps(body), headers, httplib.responses[code]


class GoogleStorageMockHttp(S3MockHttp):
    fixtures = StorageFileFixtures("google_storage")

    def _test2_test_get_object(self, method, url, body, headers):
        # test_get_object
        # Google uses a different HTTP header prefix for meta data
        body = self.fixtures.load("list_containers.xml")
        headers = {
            "content-type": "application/zip",
            "etag": '"e31208wqsdoj329jd"',
            "x-goog-meta-rabbits": "monkeys",
            "content-length": "12345",
            "last-modified": "Thu, 13 Sep 2012 07:13:22 GMT",
        }

        return httplib.OK, body, headers, httplib.responses[httplib.OK]

    def _test2_test_cont_length_get_object(self, method, url, body, headers):
        # test_get_object_object_size_not_in_content_length_header
        # Google uses a different HTTP header prefix for meta data
        body = self.fixtures.load("list_containers.xml")
        headers = {
            "content-type": "application/zip",
            "etag": '"e31208wqsdoj329jd"',
            "x-goog-meta-rabbits": "monkeys",
            "x-goog-stored-content-length": "9587",
            "last-modified": "Thu, 13 Sep 2012 07:13:22 GMT",
        }

        return httplib.OK, body, headers, httplib.responses[httplib.OK]

    def _container_path_UNAUTHORIZED(self, method, url, body, headers):
        return (
            httplib.UNAUTHORIZED,
            "",
            self.base_headers,
            httplib.responses[httplib.OK],
        )


class GoogleStorageJSONMockHttp(StorageMockHttp):
    """
    Extracts bucket and object out of requests and routes to methods of the
    forms (bucket, object, entity, and type are sanitized values
    {'-', '.', '/' are replaced with '_'}):

    _<bucket>[_<type>]
    _<bucket>_acl[_entity][_<type>]
    _<bucket>_defaultObjectAcl[_<entity>][_<type>]
    _<bucket>_<object>[_<type>]
    _<bucket>_<object>_acl[_<entity>][_<type>]

    Ugly example:
        /storage/v1/b/test-bucket/o/test-object/acl/test-entity
        with type='FOO' yields
        _test_bucket_test_object_acl_test_entity_FOO
    """

    fixtures = StorageFileFixtures("google_storage")
    base_headers = {}

    # Path regex captures bucket, object, defaultObjectAcl, and acl values.
    path_rgx = re.compile(
        r"/storage/[^/]+/b/([^/]+)"
        r"(?:/(defaultObjectAcl(?:/[^/]+)?$)|"
        r"(?:/o/(.+?))?(?:/(acl(?:/[^/]+)?))?$)"
    )

    # Permissions to use when handling requests.
    bucket_perms = google_storage.ContainerPermissions.NONE
    object_perms = google_storage.ObjectPermissions.NONE

    _FORBIDDEN = _error_helper(httplib.FORBIDDEN, base_headers)
    _NOT_FOUND = _error_helper(httplib.NOT_FOUND, base_headers)
    _PRECONDITION_FAILED = _error_helper(httplib.PRECONDITION_FAILED, base_headers)

    def _get_method_name(self, type, use_param, qs, path):
        match = self.path_rgx.match(path)
        if not match:
            raise ValueError("%s is not a valid path." % path)

        joined_groups = "_".join([g for g in match.groups() if g])
        if type:
            meth_name = "_{}_{}".format(joined_groups, type)
        else:
            meth_name = "_%s" % joined_groups
        # Return sanitized method name.
        return meth_name.replace("/", "_").replace(".", "_").replace("-", "_")

    def _response_helper(self, fixture):
        body = self.fixtures.load(fixture)
        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    ####################
    # Request handlers #
    ####################
    def _test_bucket(self, method, url, body, headers):
        """Bucket request."""
        if method != "GET":
            raise NotImplementedError("%s is not implemented." % method)

        if self.bucket_perms < google_storage.ContainerPermissions.READER:
            return self._FORBIDDEN
        else:
            return self._response_helper("get_container.json")

    def _test_bucket_acl(self, method, url, body, headers):
        """Bucket list ACL request."""
        if method != "GET":
            raise NotImplementedError("%s is not implemented." % method)

        if self.bucket_perms < google_storage.ContainerPermissions.OWNER:
            return self._FORBIDDEN
        else:
            return self._response_helper("list_container_acl.json")

    def _test_bucket_test_object(self, method, url, body, headers):
        """Object request."""
        if method != "GET":
            raise NotImplementedError("%s is not implemented." % method)

        if self.object_perms < google_storage.ObjectPermissions.READER:
            return self._FORBIDDEN
        else:
            return self._response_helper("get_object.json")

    def _test_bucket_test_object_acl(self, method, url, body, headers):
        """Object list ACL request."""
        if method != "GET":
            raise NotImplementedError("%s is not implemented." % method)

        if self.object_perms < google_storage.ObjectPermissions.OWNER:
            return self._FORBIDDEN
        else:
            return self._response_helper("list_object_acl.json")

    def _test_bucket_writecheck(self, method, url, body, headers):
        gen_match = headers.get("x-goog-if-generation-match")
        if method != "DELETE" or gen_match != "0":
            msg = "Improper write check delete strategy. method: %s, " "headers: %s" % (
                method,
                headers,
            )
            raise ValueError(msg)

        if self.bucket_perms < google_storage.ContainerPermissions.WRITER:
            return self._FORBIDDEN
        else:
            return self._PRECONDITION_FAILED


class GoogleStorageConnectionTest(GoogleTestCase):
    @mock.patch("email.utils.formatdate")
    def test_add_default_headers(self, mock_formatdate):
        mock_formatdate.return_value = TODAY
        starting_headers = {"starting": "headers"}
        project = "foo-project"

        # Modify headers when there is no project.
        conn = CONN_CLS("foo_user", "bar_key", secure=True, auth_type=GoogleAuthType.GCS_S3)
        conn.get_project = mock.Mock(return_value=None)
        headers = dict(starting_headers)
        headers["Date"] = TODAY
        self.assertEqual(conn.add_default_headers(dict(starting_headers)), headers)

        # Modify headers when there is a project.
        conn = CONN_CLS("foo_user", "bar_key", secure=True, auth_type=GoogleAuthType.GCS_S3)
        conn.get_project = mock.Mock(return_value=project)
        headers = dict(starting_headers)
        headers["Date"] = TODAY
        headers[CONN_CLS.PROJECT_ID_HEADER] = project
        self.assertEqual(conn.add_default_headers(dict(starting_headers)), headers)

    @mock.patch("libcloud.storage.drivers.s3." "BaseS3Connection.get_auth_signature")
    def test_get_s3_auth_signature(self, mock_s3_auth_sig_method):
        # Check that the S3 HMAC signature method is used.
        # Check that headers are copied and modified properly before calling
        # the signature method.
        mock_s3_auth_sig_method.return_value = "mock signature!"
        starting_params = {}
        starting_headers = {
            "Date": TODAY,
            "x-goog-foo": "X-GOOG: MAINTAIN UPPERCASE!",
            "x-Goog-bar": "Header key should be lowered",
            "Content-Type": "application/mIXED casING MAINTAINED",
            "Other": "LOWER THIS!",
        }
        modified_headers = {
            "date": TODAY,
            "content-type": "application/mIXED casING MAINTAINED",
            "x-goog-foo": "X-GOOG: MAINTAIN UPPERCASE!",
            "x-goog-bar": "Header key should be lowered",
            "other": "lower this!",
        }

        conn = CONN_CLS("foo_user", "bar_key", secure=True, auth_type=GoogleAuthType.GCS_S3)
        conn.method = "GET"
        conn.action = "/path"
        result = conn._get_s3_auth_signature(starting_params, starting_headers)
        self.assertNotEqual(starting_headers, modified_headers)
        self.assertEqual(result, "mock signature!")
        mock_s3_auth_sig_method.assert_called_once_with(
            method="GET",
            headers=modified_headers,
            params=starting_params,
            expires=None,
            secret_key="bar_key",
            path="/path",
            vendor_prefix="x-goog",
        )

    @mock.patch("libcloud.common.google.GoogleOAuth2Credential")
    def test_pre_connect_hook_oauth2(self, mock_oauth2_credential_init):
        # Check that we get the Authorization header from the OAuth2 token,
        # not from the HMAC signature method.
        # Check that the headers and pa
        mock_oauth2_credential_init.return_value = mock.Mock()
        starting_params = {"starting": "params"}
        starting_headers = {"starting": "headers"}

        conn = CONN_CLS("foo_user", "bar_key", secure=True, auth_type=GoogleAuthType.GCE)
        conn._get_s3_auth_signature = mock.Mock()
        conn.oauth2_credential = mock.Mock()
        conn.oauth2_credential.access_token = "Access_Token!"
        expected_headers = dict(starting_headers)
        expected_headers["Authorization"] = "Bearer Access_Token!"
        result = conn.pre_connect_hook(dict(starting_params), dict(starting_headers))
        self.assertEqual(result, (starting_params, expected_headers))

    def test_pre_connect_hook_hmac(self):
        # Check that we call for a HMAC signature, passing params and headers
        # Check that we properly apply the HMAC signature.
        # Check that we don't use OAuth2 credentials.
        starting_params = {"starting": "params"}
        starting_headers = {"starting": "headers"}

        def fake_hmac_method(params, headers):
            # snapshot the params and headers passed (they are modified later)
            fake_hmac_method.params_passed = copy.deepcopy(params)
            fake_hmac_method.headers_passed = copy.deepcopy(headers)
            return "fake signature!"

        conn = CONN_CLS("foo_user", "bar_key", secure=True, auth_type=GoogleAuthType.GCS_S3)
        conn._get_s3_auth_signature = fake_hmac_method
        conn.action = "GET"
        conn.method = "/foo"
        expected_headers = dict(starting_headers)
        expected_headers["Authorization"] = "{} {}:{}".format(
            google_storage.SIGNATURE_IDENTIFIER,
            "foo_user",
            "fake signature!",
        )
        result = conn.pre_connect_hook(dict(starting_params), dict(starting_headers))
        self.assertEqual(result, (dict(starting_params), expected_headers))
        self.assertEqual(fake_hmac_method.params_passed, starting_params)
        self.assertEqual(fake_hmac_method.headers_passed, starting_headers)
        self.assertIsNone(conn.oauth2_credential)


class GoogleStorageTests(S3Tests, GoogleTestCase):
    driver_type = google_storage.GoogleStorageDriver
    driver_args = STORAGE_GOOGLE_STORAGE_PARAMS
    mock_response_klass = GoogleStorageMockHttp

    def setUp(self):
        super().setUp()
        self.driver_type.jsonConnectionCls.conn_class = GoogleStorageJSONMockHttp

    def tearDown(self):
        self._remove_test_file()

    def test_billing_not_enabled(self):
        # TODO
        pass

    def test_token(self):
        # Not supported on Google Storage
        pass

    def test_get_object_object_size_in_content_length(self):
        self.mock_response_klass.type = "get_object"
        obj = self.driver.get_object(container_name="test2", object_name="test")
        self.assertEqual(obj.size, 12345)

    def test_get_object_object_size_not_in_content_length_header(self):
        self.mock_response_klass.type = "get_object"
        obj = self.driver.get_object(container_name="test2", object_name="test_cont_length")
        self.assertEqual(obj.size, 9587)

    def test_delete_permissions(self):
        mock_request = mock.Mock()
        self.driver.json_connection.request = mock_request

        # Test deleting object permissions.
        self.driver.ex_delete_permissions("bucket", "object", entity="user-foo")
        url = "/storage/v1/b/bucket/o/object/acl/user-foo"
        mock_request.assert_called_once_with(url, method="DELETE")

        # Test deleting bucket permissions.
        mock_request.reset_mock()
        self.driver.ex_delete_permissions("bucket", entity="user-foo")
        url = "/storage/v1/b/bucket/acl/user-foo"
        mock_request.assert_called_once_with(url, method="DELETE")

    def test_delete_permissions_no_entity(self):
        mock_request = mock.Mock()
        mock_get_user = mock.Mock(return_value=None)
        self.driver._get_user = mock_get_user
        self.driver.json_connection.request = mock_request

        # Test deleting permissions on an object with no entity.
        self.assertRaises(ValueError, self.driver.ex_delete_permissions, "bucket", "object")

        # Test deleting permissions on an bucket with no entity.
        self.assertRaises(ValueError, self.driver.ex_delete_permissions, "bucket")

        mock_request.assert_not_called()

        # Test deleting permissions on an object with a default entity.
        mock_get_user.return_value = "foo@foo.com"
        self.driver.ex_delete_permissions("bucket", "object")
        url = "/storage/v1/b/bucket/o/object/acl/user-foo@foo.com"
        mock_request.assert_called_once_with(url, method="DELETE")

        # Test deleting permissions on an bucket with a default entity.
        mock_request.reset_mock()
        mock_get_user.return_value = "foo@foo.com"
        self.driver.ex_delete_permissions("bucket")
        url = "/storage/v1/b/bucket/acl/user-foo@foo.com"
        mock_request.assert_called_once_with(url, method="DELETE")

    def test_get_permissions(self):
        def test_permission_config(bucket_perms, object_perms):
            GoogleStorageJSONMockHttp.bucket_perms = bucket_perms
            GoogleStorageJSONMockHttp.object_perms = object_perms

            perms = self.driver.ex_get_permissions("test-bucket", "test-object")
            self.assertEqual(perms, (bucket_perms, object_perms))

        bucket_levels = range(len(google_storage.ContainerPermissions.values))
        object_levels = range(len(google_storage.ObjectPermissions.values))
        for bucket_perms in bucket_levels:
            for object_perms in object_levels:
                test_permission_config(bucket_perms, object_perms)

    def test_set_permissions(self):
        mock_request = mock.Mock()
        self.driver.json_connection.request = mock_request

        # Test setting object permissions.
        self.driver.ex_set_permissions("bucket", "object", entity="user-foo", role="OWNER")
        url = "/storage/v1/b/bucket/o/object/acl"
        mock_request.assert_called_once_with(
            url, method="POST", data=json.dumps({"role": "OWNER", "entity": "user-foo"})
        )

        # Test setting object permissions with an ObjectPermissions value.
        mock_request.reset_mock()
        self.driver.ex_set_permissions(
            "bucket",
            "object",
            entity="user-foo",
            role=google_storage.ObjectPermissions.OWNER,
        )
        url = "/storage/v1/b/bucket/o/object/acl"
        mock_request.assert_called_once_with(
            url, method="POST", data=json.dumps({"role": "OWNER", "entity": "user-foo"})
        )

        # Test setting bucket permissions.
        mock_request.reset_mock()
        self.driver.ex_set_permissions("bucket", entity="user-foo", role="OWNER")
        url = "/storage/v1/b/bucket/acl"
        mock_request.assert_called_once_with(
            url, method="POST", data=json.dumps({"role": "OWNER", "entity": "user-foo"})
        )

        # Test setting bucket permissions with a ContainerPermissions value.
        mock_request.reset_mock()
        self.driver.ex_set_permissions(
            "bucket", entity="user-foo", role=google_storage.ContainerPermissions.OWNER
        )
        url = "/storage/v1/b/bucket/acl"
        mock_request.assert_called_once_with(
            url, method="POST", data=json.dumps({"role": "OWNER", "entity": "user-foo"})
        )

    def test_set_permissions_bad_roles(self):
        mock_request = mock.Mock()
        self.driver.json_connection.request = mock_request

        # Test forgetting a role.
        self.assertRaises(ValueError, self.driver.ex_set_permissions, "bucket", "object")
        self.assertRaises(ValueError, self.driver.ex_set_permissions, "bucket")
        mock_request.assert_not_called()

        # Test container permissions on an object.
        self.assertRaises(
            ValueError,
            self.driver.ex_set_permissions,
            "bucket",
            "object",
            role=google_storage.ContainerPermissions.OWNER,
        )
        mock_request.assert_not_called()

        # Test object permissions on a container.
        self.assertRaises(
            ValueError,
            self.driver.ex_set_permissions,
            "bucket",
            role=google_storage.ObjectPermissions.OWNER,
        )
        mock_request.assert_not_called()

    def test_set_permissions_no_entity(self):
        mock_request = mock.Mock()
        mock_get_user = mock.Mock(return_value=None)
        self.driver._get_user = mock_get_user
        self.driver.json_connection.request = mock_request

        # Test for setting object permissions with no entity.
        self.assertRaises(
            ValueError, self.driver.ex_set_permissions, "bucket", "object", role="OWNER"
        )

        # Test for setting bucket permissions with no entity.
        self.assertRaises(ValueError, self.driver.ex_set_permissions, "bucket", role="OWNER")

        mock_request.assert_not_called()

        # Test for setting object permissions with a default entity.
        mock_get_user.return_value = "foo@foo.com"
        self.driver.ex_set_permissions("bucket", "object", role="OWNER")
        url = "/storage/v1/b/bucket/o/object/acl"
        mock_request.assert_called_once_with(
            url,
            method="POST",
            data=json.dumps({"role": "OWNER", "entity": "user-foo@foo.com"}),
        )

        # Test for setting bucket permissions with a default entity.
        mock_request.reset_mock()
        mock_get_user.return_value = "foo@foo.com"
        self.driver.ex_set_permissions("bucket", role="OWNER")
        url = "/storage/v1/b/bucket/acl"
        mock_request.assert_called_once_with(
            url,
            method="POST",
            data=json.dumps({"role": "OWNER", "entity": "user-foo@foo.com"}),
        )

    def test_invalid_credentials_on_upload(self):
        self.mock_response_klass.type = "UNAUTHORIZED"
        container = Container(name="container", driver=self.driver, extra={})
        with pytest.raises(InvalidCredsError):
            self.driver.upload_object_via_stream(BytesIO(b" "), container, "path")

    def test_download_object_data_is_not_buffered_in_memory(self):
        # Test case which verifies that response.body attribute is not accessed
        # and as such, whole body response is not buffered into RAM

        # If content is consumed and response.content attribute accessed exception
        # will be thrown and test will fail

        mock_response = Mock(name="mock response")
        mock_response.headers = {}
        mock_response.status_code = 200
        msg = '"content" attribute was accessed but it shouldn\'t have been'
        type(mock_response).content = PropertyMock(
            name="mock content attribute", side_effect=Exception(msg)
        )
        mock_response.iter_content.return_value = StringIO("a" * 1000)

        self.driver.connection.connection.getresponse = Mock()
        self.driver.connection.connection.getresponse.return_value = mock_response

        container = Container(name="foo_bar_container", extra={}, driver=self.driver)
        obj = Object(
            name="foo_bar_object_NO_BUFFER",
            size=1000,
            hash=None,
            extra={},
            container=container,
            meta_data=None,
            driver=self.driver_type,
        )
        destination_path = self._file_path
        result = self.driver.download_object(
            obj=obj,
            destination_path=destination_path,
            overwrite_existing=True,
            delete_on_failure=True,
        )
        self.assertTrue(result)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.test.secrets import DNS_PARAMS_ZERIGO
from libcloud.dns.drivers.zerigo import ZerigoError, ZerigoDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures


class ZerigoTests(unittest.TestCase):
    def setUp(self):
        ZerigoDNSDriver.connectionCls.conn_class = ZerigoMockHttp
        ZerigoMockHttp.type = None
        self.driver = ZerigoDNSDriver(*DNS_PARAMS_ZERIGO)

    def test_invalid_credentials(self):
        ZerigoMockHttp.type = "INVALID_CREDS"

        try:
            list(self.driver.list_zones())
        except InvalidCredsError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 13)
        self.assertTrue(RecordType.A in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 1)
        self.assertEqual(zones[0].domain, "example.com")
        self.assertEqual(zones[0].type, "master")
        self.assertEqual(zones[0].extra["notes"], "test foo bar")

    def test_list_zones_no_results(self):
        ZerigoMockHttp.type = "NO_RESULTS"
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 0)

    def test_list_records_success(self):
        zone = self.driver.list_zones()[0]
        records = list(self.driver.list_records(zone=zone))

        self.assertEqual(len(records), 4)
        self.assertEqual(records[0].name, "www")
        self.assertEqual(records[0].type, RecordType.A)
        self.assertEqual(records[0].data, "172.16.16.1")
        self.assertEqual(records[0].extra["fqdn"], "www.example.com")
        self.assertIsNone(records[0].extra["notes"])
        self.assertIsNone(records[0].extra["priority"])

        self.assertEqual(records[1].name, "test")
        self.assertEqual(records[1].extra["ttl"], 3600)

    def test_record_with_empty_name(self):
        zone = self.driver.list_zones()[0]
        record1 = list(self.driver.list_records(zone=zone))[-1]
        record2 = list(self.driver.list_records(zone=zone))[-2]

        self.assertIsNone(record1.name)
        self.assertIsNone(record2.name)

    def test_list_records_no_results(self):
        zone = self.driver.list_zones()[0]
        ZerigoMockHttp.type = "NO_RESULTS"
        records = list(self.driver.list_records(zone=zone))
        self.assertEqual(len(records), 0)

    def test_list_records_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        ZerigoMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            list(self.driver.list_records(zone=zone))
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")
        pass

    def test_get_zone_success(self):
        zone = self.driver.get_zone(zone_id=12345678)

        self.assertEqual(zone.id, "12345678")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.extra["hostmaster"], "dnsadmin@example.com")
        self.assertEqual(zone.type, "master")

    def test_get_zone_does_not_exist(self):
        ZerigoMockHttp.type = "DOES_NOT_EXIST"

        try:
            self.driver.get_zone(zone_id="4444")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "4444")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_success(self):
        record = self.driver.get_record(zone_id="12345678", record_id="23456789")
        self.assertEqual(record.id, "23456789")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.type, RecordType.A)

    def test_get_record_zone_does_not_exist(self):
        ZerigoMockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="444", record_id="28536")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        ZerigoMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="12345678", record_id="28536")
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_success(self):
        ZerigoMockHttp.type = "CREATE_ZONE"

        zone = self.driver.create_zone(domain="foo.bar.com", type="master", ttl=None, extra=None)
        self.assertEqual(zone.id, "12345679")
        self.assertEqual(zone.domain, "foo.bar.com")

    def test_create_zone_validaton_error(self):
        ZerigoMockHttp.type = "CREATE_ZONE_VALIDATION_ERROR"

        try:
            self.driver.create_zone(domain="foo.bar.com", type="master", ttl=10, extra=None)
        except ZerigoError as e:
            self.assertEqual(len(e.errors), 2)
        else:
            self.fail("Exception was not thrown")

    def test_update_zone_success(self):
        zone = self.driver.list_zones()[0]
        updated_zone = self.driver.update_zone(zone=zone, ttl=10, extra={"notes": "bar foo"})

        self.assertEqual(zone.extra["notes"], "test foo bar")

        self.assertEqual(updated_zone.id, zone.id)
        self.assertEqual(updated_zone.domain, "example.com")
        self.assertEqual(updated_zone.type, zone.type)
        self.assertEqual(updated_zone.ttl, 10)
        self.assertEqual(updated_zone.extra["notes"], "bar foo")

    def test_update_zone_domain_cannot_be_changed(self):
        zone = self.driver.list_zones()[0]

        try:
            self.driver.update_zone(zone=zone, domain="libcloud.org")
        except LibcloudError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_record_success(self):
        zone = self.driver.list_zones()[0]

        ZerigoMockHttp.type = "CREATE_RECORD"
        record = self.driver.create_record(
            name="www", zone=zone, type=RecordType.A, data="127.0.0.1"
        )

        self.assertEqual(record.id, "23456780")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")

    def test_update_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        updated_record = self.driver.update_record(
            record=record, name="www", type=RecordType.AAAA, data="::1"
        )

        self.assertEqual(record.data, "172.16.16.1")

        self.assertEqual(updated_record.id, record.id)
        self.assertEqual(updated_record.name, "www")
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.AAAA)
        self.assertEqual(updated_record.data, "::1")

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        ZerigoMockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]

        ZerigoMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail("Exception was not thrown")


class ZerigoMockHttp(MockHttp):
    fixtures = DNSFileFixtures("zerigo")

    def _api_1_1_zones_xml_INVALID_CREDS(self, method, url, body, headers):
        body = "HTTP Basic: Access denied.\n"
        return (httplib.UNAUTHORIZED, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_xml(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {"x-query-count": "1"}, httplib.responses[httplib.OK])

    def _api_1_1_zones_xml_NO_RESULTS(self, method, url, body, headers):
        body = self.fixtures.load("list_zones_no_results.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_hosts_xml(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {"x-query-count": "1"}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_hosts_xml_NO_RESULTS(self, method, url, body, headers):
        body = self.fixtures.load("list_records_no_results.xml")
        return (httplib.OK, body, {"x-query-count": "0"}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_hosts_xml_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_xml(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_4444_xml_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_hosts_23456789_xml(self, method, url, body, headers):
        body = self.fixtures.load("get_record.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_444_xml_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_xml_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_hosts_28536_xml_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_xml_CREATE_ZONE(self, method, url, body, headers):
        body = self.fixtures.load("create_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_xml_CREATE_ZONE_VALIDATION_ERROR(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_validation_error.xml")
        return (httplib.UNPROCESSABLE_ENTITY, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_hosts_xml_CREATE_RECORD(self, method, url, body, headers):
        body = self.fixtures.load("create_record.xml")
        return (httplib.CREATED, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_zones_12345678_xml_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    def _api_1_1_hosts_23456789_xml_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = ""
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.OK])

    """
    def (self, method, url, body, headers):
        body = self.fixtures.load('.xml')
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def (self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def (self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def (self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def (self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def (self, method, url, body, headers):
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])
    """


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import CONTAINER_PARAMS_KUBERNETES
from libcloud.container.base import ContainerImage
from libcloud.test.file_fixtures import ContainerFileFixtures
from libcloud.test.common.test_kubernetes import KubernetesAuthTestCaseMixin
from libcloud.container.drivers.kubernetes import (
    KubernetesContainerDriver,
    to_n_cpus,
    to_cpu_str,
    to_n_bytes,
    sum_resources,
    to_memory_str,
)


class KubernetesContainerDriverTestCase(unittest.TestCase, KubernetesAuthTestCaseMixin):
    driver_cls = KubernetesContainerDriver

    def setUp(self):
        KubernetesContainerDriver.connectionCls.conn_class = KubernetesMockHttp
        KubernetesMockHttp.type = None
        KubernetesMockHttp.use_param = "a"
        self.driver = KubernetesContainerDriver(*CONTAINER_PARAMS_KUBERNETES)

    def test_list_containers(self):
        containers = self.driver.list_containers()
        self.assertEqual(len(containers), 1)
        self.assertEqual(
            containers[0].id,
            "docker://3c48b5cda79bce4c8866f02a3b96a024edb8f660d10e7d1755e9ced49ef47b36",
        )
        self.assertEqual(containers[0].name, "hello-world")

    def test_deploy_container(self):
        image = ContainerImage(
            id=None, name="hello-world", path=None, driver=self.driver, version=None
        )
        container = self.driver.deploy_container("hello-world", image=image)
        self.assertEqual(container.name, "hello-world")

    def test_get_container(self):
        container = self.driver.get_container(
            "docker://3c48b5cda79bce4c8866f02a3b96a024edb8f660d10e7d1755e9ced49ef47b36"
        )
        assert (
            container.id
            == "docker://3c48b5cda79bce4c8866f02a3b96a024edb8f660d10e7d1755e9ced49ef47b36"
        )

    def test_list_namespaces(self):
        namespaces = self.driver.list_namespaces()
        self.assertEqual(len(namespaces), 2)
        self.assertEqual(namespaces[0].id, "default")
        self.assertEqual(namespaces[0].name, "default")

    def test_get_namespace(self):
        namespace = self.driver.get_namespace("default")
        self.assertEqual(namespace.id, "default")
        self.assertEqual(namespace.name, "default")

    def test_create_namespace(self):
        namespace = self.driver.create_namespace("test")
        self.assertEqual(namespace.id, "test")
        self.assertEqual(namespace.name, "test")

    def test_delete_namespace(self):
        namespace = self.driver.get_namespace("default")
        result = self.driver.delete_namespace(namespace)
        self.assertTrue(result)

    def test_list_pods(self):
        pods = self.driver.ex_list_pods()
        self.assertEqual(len(pods), 1)
        self.assertEqual(pods[0].id, "1fad5411-b9af-11e5-8701-0050568157ec")
        self.assertEqual(pods[0].name, "hello-world")

    def test_destroy_pod(self):
        result = self.driver.ex_destroy_pod("default", "default")
        self.assertTrue(result)

    def test_list_nodes(self):
        nodes = self.driver.ex_list_nodes()
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, "45949cbb-b99d-11e5-8d53-0050568157ec")
        self.assertEqual(nodes[0].name, "127.0.0.1")

    def test_destroy_node(self):
        result = self.driver.ex_destroy_node("127.0.0.1")
        self.assertTrue(result)

    def test_get_version(self):
        version = self.driver.ex_get_version()
        self.assertEqual(version, "v1.20.8-gke.900")

    def test_list_nodes_metrics(self):
        nodes_metrics = self.driver.ex_list_nodes_metrics()
        self.assertEqual(len(nodes_metrics), 1)
        self.assertEqual(
            nodes_metrics[0]["metadata"]["name"],
            "gke-cluster-3-default-pool-76fd57f7-l83v",
        )

    def test_list_pods_metrics(self):
        pods_metrics = self.driver.ex_list_pods_metrics()
        self.assertEqual(len(pods_metrics), 10)
        self.assertEqual(pods_metrics[0]["metadata"]["name"], "gke-metrics-agent-sfjzj")
        self.assertEqual(
            pods_metrics[1]["metadata"]["name"],
            "stackdriver-metadata-agent-cluster-level-849ff68b6d-fphxl",
        )
        self.assertEqual(pods_metrics[2]["metadata"]["name"], "event-exporter-gke-67986489c8-g47rz")

    def test_list_services(self):
        services = self.driver.ex_list_services()
        self.assertEqual(len(services), 4)
        self.assertEqual(services[0]["metadata"]["name"], "kubernetes")
        self.assertEqual(services[1]["metadata"]["name"], "default-http-backend")
        self.assertEqual(services[2]["metadata"]["name"], "kube-dns")

    def test_list_deployments(self):
        deployments = self.driver.ex_list_deployments()
        self.assertEqual(len(deployments), 7)
        deployment = deployments[0]
        self.assertEqual(deployment.id, "aea45586-9a4a-4a01-805c-719f431316c0")
        self.assertEqual(deployment.name, "event-exporter-gke")
        self.assertEqual(deployment.namespace, "kube-system")
        for deployment in deployments:
            self.assertIsInstance(deployment.replicas, int)
            self.assertIsInstance(deployment.selector, dict)

    def test_to_n_bytes(self):
        memory = "0"
        self.assertEqual(to_n_bytes(memory), 0)
        memory = "1000Ki"
        self.assertEqual(to_n_bytes(memory), 1_024_000)
        memory = "100K"
        self.assertEqual(to_n_bytes(memory), 100_000)
        memory = "512Mi"
        self.assertEqual(to_n_bytes(memory), 536_870_912)
        memory = "900M"
        self.assertEqual(to_n_bytes(memory), 900_000_000)
        memory = "10Gi"
        self.assertEqual(to_n_bytes(memory), 10_737_418_240)
        memory = "10G"
        self.assertEqual(to_n_bytes(memory), 10_000_000_000)

    def test_to_memory_str(self):
        memory = 0
        self.assertEqual(to_memory_str(memory), "0K")
        memory = 1_024_000
        self.assertEqual(to_memory_str(memory), "1000Ki")
        memory = 100_000
        self.assertEqual(to_memory_str(memory), "100K")
        memory = 536_870_912
        self.assertEqual(to_memory_str(memory), "512Mi")
        memory = 900_000_000
        self.assertEqual(to_memory_str(memory), "900M")
        memory = 10_737_418_240
        self.assertEqual(to_memory_str(memory), "10Gi")
        memory = 10_000_000_000
        self.assertEqual(to_memory_str(memory), "10G")

    def test_to_cpu_str(self):
        cpu = 0
        self.assertEqual(to_cpu_str(cpu), "0")
        cpu = 0.5
        self.assertEqual(to_cpu_str(cpu), "500m")
        cpu = 2
        self.assertEqual(to_cpu_str(cpu), "2000m")
        cpu = 0.000001
        self.assertEqual(to_cpu_str(cpu), "1u")
        cpu = 0.0005
        self.assertEqual(to_cpu_str(cpu), "500u")
        cpu = 0.000000001
        self.assertEqual(to_cpu_str(cpu), "1n")
        cpu = 0.0000005
        self.assertEqual(to_cpu_str(cpu), "500n")

    def test_to_n_cpus(self):
        cpu = "0m"
        self.assertEqual(to_n_cpus(cpu), 0)
        cpu = "2"
        self.assertEqual(to_n_cpus(cpu), 2)
        cpu = "500m"
        self.assertEqual(to_n_cpus(cpu), 0.5)
        cpu = "500m"
        self.assertEqual(to_n_cpus(cpu), 0.5)
        cpu = "2000m"
        self.assertEqual(to_n_cpus(cpu), 2)
        cpu = "1u"
        self.assertEqual(to_n_cpus(cpu), 0.000001)
        cpu = "500u"
        self.assertEqual(to_n_cpus(cpu), 0.0005)
        cpu = "1n"
        self.assertEqual(to_n_cpus(cpu), 0.000000001)
        cpu = "500n"
        self.assertEqual(to_n_cpus(cpu), 0.0000005)

    def test_sum_resources(self):
        resource_1 = {"cpu": "1", "memory": "1000Mi"}
        resource_2 = {"cpu": "2", "memory": "2000Mi"}
        self.assertDictEqual(
            sum_resources(resource_1, resource_2),
            {"cpu": "3000m", "memory": "3000Mi"},
        )
        resource_3 = {"cpu": "1500m", "memory": "1Gi"}
        self.assertDictEqual(
            sum_resources(resource_1, resource_2, resource_3),
            {"cpu": "4500m", "memory": "4024Mi"},
        )


class KubernetesMockHttp(MockHttp):
    fixtures = ContainerFileFixtures("kubernetes")

    def _api_v1_pods(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_pods.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_namespaces(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_namespaces.json")
        elif method == "POST":
            body = self.fixtures.load("_api_v1_namespaces_test.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_namespaces_default(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_namespaces_default.json")
        elif method == "DELETE":
            body = self.fixtures.load("_api_v1_namespaces_default_DELETE.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_namespaces_default_pods(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_namespaces_default_pods.json")
        elif method == "POST":
            body = self.fixtures.load("_api_v1_namespaces_default_pods_POST.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_namespaces_default_pods_default(self, method, url, body, headers):
        if method == "DELETE":
            body = None
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_nodes(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_nodes.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_nodes_127_0_0_1(self, method, url, body, headers):
        if method == "DELETE":
            body = None
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _api_v1_services(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_api_v1_services.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _version(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_version.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _apis_metrics_k8s_io_v1beta1_nodes(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_apis_metrics_k8s_io_v1beta1_nodes.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _apis_metrics_k8s_io_v1beta1_pods(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("_apis_metrics_k8s_io_v1beta1_pods.json")
        else:
            raise AssertionError("Unsupported method")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _apis_apps_v1_deployments(self, method, url, body, headers):
        body = self.fixtures.load("_apis_apps_v1_deployments.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import CLOUDSCALE_PARAMS
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.cloudscale import CloudscaleNodeDriver

try:
    import simplejson as json
except ImportError:
    import json  # NOQA


class CloudscaleTests(LibcloudTestCase):
    def setUp(self):
        CloudscaleNodeDriver.connectionCls.conn_class = CloudscaleMockHttp
        self.driver = CloudscaleNodeDriver(*CLOUDSCALE_PARAMS)

    def test_list_images_success(self):
        images = self.driver.list_images()
        (image,) = images
        self.assertTrue(image.id is not None)
        self.assertTrue(image.name is not None)

    def test_list_sizes_success(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(len(sizes), 2)

        size = sizes[0]
        self.assertTrue(size.id is not None)
        self.assertEqual(size.name, "Flex-2")
        self.assertEqual(size.ram, 2048)

        size = sizes[1]
        self.assertTrue(size.id is not None)
        self.assertEqual(size.name, "Flex-4")
        self.assertEqual(size.ram, 4096)

    def test_list_locations_not_existing(self):
        # assertRaises doesn't exist in Python 2.5?!
        try:
            self.driver.list_locations()
        except NotImplementedError:
            pass
        else:
            assert False, "Did not raise the wished error."

    def test_list_nodes_success(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, "47cec963-fcd2-482f-bdb6-24461b2d47b1")
        self.assertEqual(nodes[0].public_ips, ["185.98.122.176", "2a06:c01:1:1902::7ab0:176"])

    def test_reboot_node_success(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.reboot_node(node)
        self.assertTrue(result)

    def test_create_node_success(self):
        test_size = self.driver.list_sizes()[0]
        test_image = self.driver.list_images()[0]
        created_node = self.driver.create_node("node-name", test_size, test_image)
        self.assertEqual(created_node.id, "47cec963-fcd2-482f-bdb6-24461b2d47b1")

    def test_destroy_node_success(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.destroy_node(node)
        self.assertTrue(result)


class CloudscaleMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("cloudscale")

    def _v1_images(self, method, url, body, headers):
        body = self.fixtures.load("list_images.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_flavors(self, method, url, body, headers):
        body = self.fixtures.load("list_sizes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v1_servers(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_nodes.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        else:
            body = self.fixtures.load("create_node.json")
            response = httplib.responses[httplib.CREATED]
            return (httplib.CREATED, body, {}, response)

    def _v1_servers_47cec963_fcd2_482f_bdb6_24461b2d47b1(self, method, url, body, headers):
        assert method == "DELETE"
        return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.NO_CONTENT])

    def _v1_servers_47cec963_fcd2_482f_bdb6_24461b2d47b1_reboot(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RimuHosting Driver
"""
try:
    import simplejson as json
except ImportError:
    import json

from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.compute.types import Provider, NodeState

API_CONTEXT = "/r"
API_HOST = "rimuhosting.com"


class RimuHostingException(Exception):
    """
    Exception class for RimuHosting driver
    """

    def __str__(self):
        # pylint: disable=unsubscriptable-object
        return self.args[0]

    def __repr__(self):
        # pylint: disable=unsubscriptable-object
        return "<RimuHostingException '%s'>" % (self.args[0])


class RimuHostingResponse(JsonResponse):
    """
    Response Class for RimuHosting driver
    """

    def success(self):
        if self.status == 403:
            raise InvalidCredsError()
        return True

    def parse_body(self):
        try:
            js = super().parse_body()
            keys = list(js.keys())
            if js[keys[0]]["response_type"] == "ERROR":
                raise RimuHostingException(js[keys[0]]["human_readable_message"])
            return js[keys[0]]
        except KeyError:
            raise RimuHostingException("Could not parse body: %s" % (self.body))


class RimuHostingConnection(ConnectionKey):
    """
    Connection class for the RimuHosting driver
    """

    api_context = API_CONTEXT
    host = API_HOST
    port = 443
    responseCls = RimuHostingResponse

    def __init__(self, key, secure=True, retry_delay=None, backoff=None, timeout=None):
        # override __init__ so that we can set secure of False for testing
        ConnectionKey.__init__(
            self, key, secure, timeout=timeout, retry_delay=retry_delay, backoff=backoff
        )

    def add_default_headers(self, headers):
        # We want JSON back from the server. Could be application/xml
        # (but JSON is better).
        headers["Accept"] = "application/json"
        # Must encode all data as json, or override this header.
        headers["Content-Type"] = "application/json"

        headers["Authorization"] = "rimuhosting apikey=%s" % (self.key)
        return headers

    def request(self, action, params=None, data="", headers=None, method="GET"):
        if not headers:
            headers = {}
        if not params:
            params = {}
        # Override this method to prepend the api_context
        return ConnectionKey.request(self, self.api_context + action, params, data, headers, method)


class RimuHostingNodeDriver(NodeDriver):
    """
    RimuHosting node driver
    """

    type = Provider.RIMUHOSTING
    name = "RimuHosting"
    website = "http://rimuhosting.com/"
    connectionCls = RimuHostingConnection
    features = {"create_node": ["password"]}

    def __init__(self, key, host=API_HOST, port=443, api_context=API_CONTEXT, secure=True):
        """
        :param    key: API key (required)
        :type     key: ``str``

        :param    host: hostname for connection
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :param    api_context: Optional API context.
        :type     api_context: ``str``

        :param    secure: Whether to use HTTPS or HTTP.
        :type     secure: ``bool``

        :rtype: ``None``
        """
        # Pass in some extra vars so that
        self.key = key
        self.secure = secure
        self.connection = self.connectionCls(key, secure)
        self.connection.host = host
        self.connection.api_context = api_context
        self.connection.port = port
        self.connection.driver = self
        self.connection.connect()

    def _order_uri(self, node, resource):
        # Returns the order uri with its resource appended.
        return "/orders/{}/{}".format(node.id, resource)

    # TODO: Get the node state.
    def _to_node(self, order):
        n = Node(
            id=order["slug"],
            name=order["domain_name"],
            state=NodeState.RUNNING,
            public_ips=(
                [order["allocated_ips"]["primary_ip"]] + order["allocated_ips"]["secondary_ips"]
            ),
            private_ips=[],
            driver=self.connection.driver,
            extra={
                "order_oid": order["order_oid"],
                "monthly_recurring_fee": order.get("billing_info").get("monthly_recurring_fee"),
            },
        )
        return n

    def _to_size(self, plan):
        return NodeSize(
            id=plan["pricing_plan_code"],
            name=plan["pricing_plan_description"],
            ram=plan["minimum_memory_mb"],
            disk=plan["minimum_disk_gb"],
            bandwidth=plan["minimum_data_transfer_allowance_gb"],
            price=plan["monthly_recurring_amt"]["amt_usd"],
            driver=self.connection.driver,
        )

    def _to_image(self, image):
        return NodeImage(
            id=image["distro_code"],
            name=image["distro_description"],
            driver=self.connection.driver,
        )

    def list_sizes(self, location=None):
        # Returns a list of sizes (aka plans)
        # Get plans. Note this is really just for libcloud.
        # We are happy with any size.
        if location is None:
            location = ""
        else:
            location = ";dc_location=%s" % (location.id)

        res = self.connection.request("/pricing-plans;server-type=VPS%s" % (location)).object
        return list(map(lambda x: self._to_size(x), res["pricing_plan_infos"]))

    def list_nodes(self):
        # Returns a list of Nodes
        # Will only include active ones.
        res = self.connection.request("/orders;include_inactive=N").object
        return list(map(lambda x: self._to_node(x), res["about_orders"]))

    def list_images(self, location=None):
        # Get all base images.
        # TODO: add other image sources. (Such as a backup of a VPS)
        # All Images are available for use at all locations
        res = self.connection.request("/distributions").object
        return list(map(lambda x: self._to_image(x), res["distro_infos"]))

    def reboot_node(self, node):
        # Reboot
        # PUT the state of RESTARTING to restart a VPS.
        # All data is encoded as JSON
        data = {"reboot_request": {"running_state": "RESTARTING"}}
        uri = self._order_uri(node, "vps/running-state")
        self.connection.request(uri, data=json.dumps(data), method="PUT")
        # XXX check that the response was actually successful
        return True

    def destroy_node(self, node):
        # Shutdown a VPS.
        uri = self._order_uri(node, "vps")
        self.connection.request(uri, method="DELETE")
        # XXX check that the response was actually successful
        return True

    def create_node(
        self,
        name,
        size,
        image,
        auth=None,
        ex_billing_oid=None,
        ex_host_server_oid=None,
        ex_vps_order_oid_to_clone=None,
        ex_num_ips=1,
        ex_extra_ip_reason=None,
        ex_memory_mb=None,
        ex_disk_space_mb=None,
        ex_disk_space_2_mb=None,
        ex_control_panel=None,
    ):
        """Creates a RimuHosting instance

        @inherits: :class:`NodeDriver.create_node`

        :keyword    name: Must be a FQDN. e.g example.com.
        :type       name: ``str``

        :keyword    ex_billing_oid: If not set,
                                    a billing method is automatically picked.
        :type       ex_billing_oid: ``str``

        :keyword    ex_host_server_oid: The host server to set the VPS up on.
        :type       ex_host_server_oid: ``str``

        :keyword    ex_vps_order_oid_to_clone: Clone another VPS to use as
                                                the image for the new VPS.
        :type       ex_vps_order_oid_to_clone: ``str``

        :keyword    ex_num_ips: Number of IPs to allocate. Defaults to 1.
        :type       ex_num_ips: ``int``

        :keyword    ex_extra_ip_reason: Reason for needing the extra IPs.
        :type       ex_extra_ip_reason: ``str``

        :keyword    ex_memory_mb: Memory to allocate to the VPS.
        :type       ex_memory_mb: ``int``

        :keyword    ex_disk_space_mb: Diskspace to allocate to the VPS.
            Defaults to 4096 (4GB).
        :type       ex_disk_space_mb: ``int``

        :keyword    ex_disk_space_2_mb: Secondary disk size allocation.
                                        Disabled by default.
        :type       ex_disk_space_2_mb: ``int``

        :keyword    ex_control_panel: Control panel to install on the VPS.
        :type       ex_control_panel: ``str``
        """
        # Note we don't do much error checking in this because we
        # expect the API to error out if there is a problem.
        data = {
            "instantiation_options": {"domain_name": name, "distro": image.id},
            "pricing_plan_code": size.id,
            "vps_parameters": {},
        }

        if ex_control_panel:
            data["instantiation_options"]["control_panel"] = ex_control_panel

        auth = self._get_and_check_auth(auth)
        data["instantiation_options"]["password"] = auth.password

        if ex_billing_oid:
            # TODO check for valid oid.
            data["billing_oid"] = ex_billing_oid

        if ex_host_server_oid:
            data["host_server_oid"] = ex_host_server_oid

        if ex_vps_order_oid_to_clone:
            data["vps_order_oid_to_clone"] = ex_vps_order_oid_to_clone

        if ex_num_ips and int(ex_num_ips) > 1:
            if not ex_extra_ip_reason:
                raise RimuHostingException("Need an reason for having an extra IP")
            else:
                if "ip_request" not in data:
                    data["ip_request"] = {}
                data["ip_request"]["num_ips"] = int("ex_num_ips")
                data["ip_request"]["extra_ip_reason"] = ex_extra_ip_reason

        if ex_memory_mb:
            data["vps_parameters"]["memory_mb"] = ex_memory_mb

        if ex_disk_space_mb:
            data["vps_parameters"]["disk_space_mb"] = ex_disk_space_mb

        if ex_disk_space_2_mb:
            data["vps_parameters"]["disk_space_2_mb"] = ex_disk_space_2_mb

        # Don't send empty 'vps_parameters' attribute
        if not data["vps_parameters"]:
            del data["vps_parameters"]

        res = self.connection.request(
            "/orders/new-vps", method="POST", data=json.dumps({"new-vps": data})
        ).object
        node = self._to_node(res["about_order"])
        node.extra["password"] = res["new_order_request"]["instantiation_options"]["password"]
        return node

    def list_locations(self):
        return [
            NodeLocation("DCAUCKLAND", "RimuHosting Auckland", "NZ", self),
            NodeLocation("DCDALLAS", "RimuHosting Dallas", "US", self),
            NodeLocation("DCLONDON", "RimuHosting London", "GB", self),
            NodeLocation("DCSYDNEY", "RimuHosting Sydney", "AU", self),
        ]



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Upcloud node driver
"""
import json
import base64

from libcloud.utils.py3 import b, httplib
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeState, NodeDriver, NodeLocation
from libcloud.compute.types import Provider
from libcloud.common.upcloud import (
    PlanPrice,
    UpcloudNodeDestroyer,
    UpcloudNodeOperations,
    UpcloudCreateNodeRequestBody,
)


class UpcloudResponse(JsonResponse):
    """
    Response class for UpcloudDriver
    """

    def success(self):
        if self.status == httplib.NO_CONTENT:
            return True
        return super().success()

    def parse_error(self):
        data = self.parse_body()
        if self.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError(value=data["error"]["error_message"])
        return data


class UpcloudConnection(ConnectionUserAndKey):
    """
    Connection class for UpcloudDriver
    """

    host = "api.upcloud.com"
    responseCls = UpcloudResponse

    def add_default_headers(self, headers):
        """Adds headers that are needed for all requests"""
        headers["Authorization"] = self._basic_auth()
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"
        return headers

    def _basic_auth(self):
        """Constructs basic auth header content string"""
        credentials = b("{}:{}".format(self.user_id, self.key))
        credentials = base64.b64encode(credentials)
        return "Basic {}".format(credentials.decode("ascii"))


class UpcloudDriver(NodeDriver):
    """
    Upcloud node driver

    :keyword    username: Username required for authentication
    :type       username: ``str``

    :keyword    password: Password required for authentication
    :type       password: ``str``
    """

    type = Provider.UPCLOUD
    name = "Upcloud"
    website = "https://www.upcloud.com"
    connectionCls = UpcloudConnection
    features = {"create_node": ["ssh_key", "generates_password"]}

    NODE_STATE_MAP = {
        "started": NodeState.RUNNING,
        "stopped": NodeState.STOPPED,
        "maintenance": NodeState.RECONFIGURING,
        "error": NodeState.ERROR,
    }

    def __init__(self, username, password, **kwargs):
        super().__init__(key=username, secret=password, **kwargs)

    def list_locations(self):
        """
        List available locations for deployment

        :rtype: ``list`` of :class:`NodeLocation`
        """
        response = self.connection.request("1.2/zone")
        return self._to_node_locations(response.object["zones"]["zone"])

    def list_sizes(self, location=None):
        """
        List available plans

        :param location: Location of the deployment. Price depends on
        location. lf location is not given or price not found for
        location, price will be None (optional)
        :type location: :class:`.NodeLocation`

        :rtype: ``list`` of :class:`NodeSize`
        """
        prices_response = self.connection.request("1.2/price")
        response = self.connection.request("1.2/plan")
        return self._to_node_sizes(
            response.object["plans"]["plan"],
            prices_response.object["prices"]["zone"],
            location,
        )

    def list_images(self):
        """
        List available distributions.

        :rtype: ``list`` of :class:`NodeImage`
        """
        response = self.connection.request("1.2/storage/template")
        obj = response.object
        response = self.connection.request("1.2/storage/cdrom")
        storage = response.object["storages"]["storage"]
        obj["storages"]["storage"].extend(storage)
        return self._to_node_images(obj["storages"]["storage"])

    def create_node(
        self,
        name,
        size,
        image,
        location,
        auth=None,
        ex_hostname="localhost",
        ex_username="root",
    ):
        """
        Creates instance to upcloud.

        If auth is not given then password will be generated.

        :param name:   String with a name for this new node (required)
        :type name:   ``str``

        :param size:   The size of resources allocated to this node.
                            (required)
        :type size:   :class:`.NodeSize`

        :param image:  OS Image to boot on node. (required)
        :type image:  :class:`.NodeImage`

        :param location: Which data center to create a node in. If empty,
                              undefined behavior will be selected. (optional)
        :type location: :class:`.NodeLocation`

        :param auth:   Initial authentication information for the node
                            (optional)
        :type auth:   :class:`.NodeAuthSSHKey`

        :param ex_hostname: Hostname. Default is 'localhost'. (optional)
        :type ex_hostname: ``str``

        :param ex_username: User's username, which is created.
                            Default is 'root'. (optional)
        :type ex_username: ``str``

        :return: The newly created node.
        :rtype: :class:`.Node`
        """
        body = UpcloudCreateNodeRequestBody(
            name=name,
            size=size,
            image=image,
            location=location,
            auth=auth,
            ex_hostname=ex_hostname,
            ex_username=ex_username,
        )
        response = self.connection.request("1.2/server", method="POST", data=body.to_json())
        server = response.object["server"]
        # Upcloud server's are in maintenance state when going
        # from state to other, it is safe to assume STARTING state
        return self._to_node(server, state=NodeState.STARTING)

    def list_nodes(self):
        """
        List nodes

        :return: List of node objects
        :rtype: ``list`` of :class:`Node`
        """
        servers = []
        for nid in self._node_ids():
            response = self.connection.request("1.2/server/{}".format(nid))
            servers.append(response.object["server"])
        return self._to_nodes(servers)

    def reboot_node(self, node):
        """
        Reboot the given node

        :param      node: the node to reboot
        :type       node: :class:`Node`

        :rtype: ``bool``
        """
        body = {"restart_server": {"stop_type": "hard"}}
        self.connection.request(
            "1.2/server/{}/restart".format(node.id),
            method="POST",
            data=json.dumps(body),
        )
        return True

    def destroy_node(self, node):
        """
        Destroy the given node

        The disk resources, attached to node,  will not be removed.

        :param       node: the node to destroy
        :type        node: :class:`Node`

        :rtype: ``bool``
        """

        operations = UpcloudNodeOperations(self.connection)
        destroyer = UpcloudNodeDestroyer(operations)
        return destroyer.destroy_node(node.id)

    def _node_ids(self):
        """
        Returns list of server uids currently on upcloud
        """
        response = self.connection.request("1.2/server")
        servers = response.object["servers"]["server"]
        return [server["uuid"] for server in servers]

    def _to_nodes(self, servers):
        return [self._to_node(server) for server in servers]

    def _to_node(self, server, state=None):
        ip_addresses = server["ip_addresses"]["ip_address"]
        public_ips = [ip["address"] for ip in ip_addresses if ip["access"] == "public"]
        private_ips = [ip["address"] for ip in ip_addresses if ip["access"] == "private"]

        extra = {"vnc_password": server["vnc_password"]}
        if "password" in server:
            extra["password"] = server["password"]
        return Node(
            id=server["uuid"],
            name=server["title"],
            state=state or self.NODE_STATE_MAP[server["state"]],
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self,
            extra=extra,
        )

    def _to_node_locations(self, zones):
        return [self._construct_node_location(zone) for zone in zones]

    def _construct_node_location(self, zone):
        return NodeLocation(
            id=zone["id"],
            name=zone["description"],
            country=self._parse_country(zone["id"]),
            driver=self,
        )

    def _parse_country(self, zone_id):
        """Parses the country information out of zone_id.
        Zone_id format [country]_[city][number], like fi_hel1"""
        return zone_id.split("-")[0].upper()

    def _to_node_sizes(self, plans, prices, location):
        plan_price = PlanPrice(prices)
        return [self._to_node_size(plan, plan_price, location) for plan in plans]

    def _to_node_size(self, plan, plan_price, location):
        extra = self._copy_dict(("core_number", "storage_tier"), plan)
        return NodeSize(
            id=plan["name"],
            name=plan["name"],
            ram=plan["memory_amount"],
            disk=plan["storage_size"],
            bandwidth=plan["public_traffic_out"],
            price=plan_price.get_price(plan["name"], location),
            driver=self,
            extra=extra,
        )

    def _to_node_images(self, images):
        return [self._construct_node_image(image) for image in images]

    def _construct_node_image(self, image):
        extra = self._copy_dict(("access", "license", "size", "state", "type"), image)
        return NodeImage(id=image["uuid"], name=image["title"], driver=self, extra=extra)

    def _copy_dict(self, keys, d):
        extra = {}
        for key in keys:
            extra[key] = d[key]
        return extra



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import json
import unittest

from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    RecordError,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_GANDI_LIVE
from libcloud.common.gandi_live import JsonParseError, GandiLiveBaseError, InvalidRequestError
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.gandi_live import GandiLiveDNSDriver
from libcloud.test.common.test_gandi_live import BaseGandiLiveMockHttp


class GandiLiveTests(unittest.TestCase):
    def setUp(self):
        GandiLiveDNSDriver.connectionCls.conn_class = GandiLiveMockHttp
        GandiLiveMockHttp.type = None
        self.driver = GandiLiveDNSDriver(*DNS_GANDI_LIVE)
        self.test_zone = Zone(
            id="example.com",
            type="master",
            ttl=None,
            domain="example.com",
            extra={"zone_uuid": "a53re"},
            driver=self,
        )
        self.test_bad_zone = Zone(
            id="badexample.com",
            type="master",
            ttl=None,
            domain="badexample.com",
            extra={"zone_uuid": "a53rf"},
            driver=self,
        )
        self.test_record = Record(
            id="A:bob",
            type=RecordType.A,
            name="bob",
            zone=self.test_zone,
            data="127.0.0.1",
            driver=self,
            extra={},
        )
        self.test_bad_record = Record(
            id="A:jane",
            type=RecordType.A,
            name="jane",
            zone=self.test_bad_zone,
            data="127.0.0.1",
            driver=self,
            extra={},
        )

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)
        zone = zones[0]
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "example.com")
        self.assertIsNone(zone.ttl)
        zone = zones[1]
        self.assertEqual(zone.id, "example.net")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "example.net")
        self.assertIsNone(zone.ttl)

    def test_create_zone(self):
        zone = self.driver.create_zone("example.org", extra={"name": "Example"})
        self.assertEqual(zone.id, "example.org")
        self.assertEqual(zone.domain, "example.org")
        self.assertEqual(zone.extra["zone_uuid"], "54321")

    def test_create_zone_without_name(self):
        zone = self.driver.create_zone("example.org")
        self.assertEqual(zone.id, "example.org")
        self.assertEqual(zone.domain, "example.org")
        self.assertEqual(zone.extra["zone_uuid"], "54321")

    def test_get_zone(self):
        zone = self.driver.get_zone("example.com")
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "example.com")
        self.assertIsNone(zone.ttl)

    def test_list_records(self):
        records = self.driver.list_records(self.test_zone)
        self.assertEqual(len(records), 3)
        record = records[0]
        self.assertEqual(record.id, "A:@")
        self.assertEqual(record.name, "@")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")
        record = records[1]
        self.assertEqual(record.id, "CNAME:www")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.type, RecordType.CNAME)
        self.assertEqual(record.data, "bob.example.com.")
        record = records[2]
        self.assertEqual(record.id, "A:bob")
        self.assertEqual(record.name, "bob")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.1.1")

    def test_get_record(self):
        record = self.driver.get_record(self.test_zone.id, "A:bob")
        self.assertEqual(record.id, "A:bob")
        self.assertEqual(record.name, "bob")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.1.1")

    def test_create_record(self):
        record = self.driver.create_record(
            "alice", self.test_zone, "AAAA", "::1", extra={"ttl": 400}
        )
        self.assertEqual(record.id, "AAAA:alice")
        self.assertEqual(record.name, "alice")
        self.assertEqual(record.type, RecordType.AAAA)
        self.assertEqual(record.data, "::1")

    def test_create_record_doesnt_throw_if_ttl_is_not_provided(self):
        record = self.driver.create_record("alice", self.test_zone, "AAAA", "::1")
        self.assertEqual(record.id, "AAAA:alice")
        self.assertEqual(record.name, "alice")
        self.assertEqual(record.type, RecordType.AAAA)
        self.assertEqual(record.data, "::1")

    def test_bad_record_validation(self):
        with self.assertRaises(RecordError) as ctx:
            self.driver.create_record(
                "alice", self.test_zone, "AAAA", "1" * 1025, extra={"ttl": 400}
            )
        self.assertTrue("Record data must be" in str(ctx.exception))
        with self.assertRaises(RecordError) as ctx:
            self.driver.create_record("alice", self.test_zone, "AAAA", "::1", extra={"ttl": 10})
        self.assertTrue("TTL must be at least" in str(ctx.exception))
        with self.assertRaises(RecordError) as ctx:
            self.driver.create_record(
                "alice", self.test_zone, "AAAA", "::1", extra={"ttl": 31 * 24 * 60 * 60}
            )
        self.assertTrue("TTL must not exceed" in str(ctx.exception))

    def test_update_record(self):
        record = self.driver.update_record(
            self.test_record, "bob", RecordType.A, "192.168.0.2", {"ttl": 500}
        )
        self.assertEqual(record.id, "A:bob")
        self.assertEqual(record.name, "bob")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "192.168.0.2")

    def test_delete_record(self):
        success = self.driver.delete_record(self.test_record)
        self.assertTrue(success)

    def test_export_bind(self):
        bind_export = self.driver.export_zone_to_bind_format(self.test_zone)
        bind_lines = bind_export.decode("utf8").split("\n")
        self.assertEqual(bind_lines[0], "@ 10800 IN A 127.0.0.1")

    def test_bad_json_response(self):
        with self.assertRaises(JsonParseError):
            self.driver.get_zone("badexample.com")

    def test_no_record_found(self):
        with self.assertRaises(RecordDoesNotExistError):
            self.driver.get_record(self.test_zone.id, "A:none")

    def test_record_already_exists(self):
        with self.assertRaises(RecordAlreadyExistsError):
            self.driver.create_record(
                "bob", self.test_bad_zone, "A", "127.0.0.1", extra={"ttl": 400}
            )

    def test_no_zone_found(self):
        with self.assertRaises(ZoneDoesNotExistError):
            self.driver.get_zone("nosuchzone.com")

    def test_zone_already_exists(self):
        with self.assertRaises(ZoneAlreadyExistsError):
            self.driver.create_zone("badexample.com")

    # This is somewhat spurious - the driver should be preventing any invalid
    # request error from arising.
    def test_suberrors(self):
        with self.assertRaises(InvalidRequestError) as ctx:
            self.driver.update_record(
                self.test_bad_record, "jane", RecordType.A, "192.168.0.2", {"ttl": 500}
            )
        self.assertTrue("is not a foo" in str(ctx.exception))

    def test_other_error(self):
        with self.assertRaises(GandiLiveBaseError):
            self.driver.list_records(self.test_bad_zone)

    def test_mx_record(self):
        record = self.driver.get_record(self.test_zone.id, "MX:lists")
        self.assertEqual(record.extra["priority"], "10")
        self.assertTrue("_other_records" in record.extra)
        other_record = record.extra["_other_records"][0]
        self.assertEqual(other_record["extra"]["priority"], "20")

    def test_ex_create_multivalue_record(self):
        records = self.driver.ex_create_multi_value_record(
            "alice", self.test_zone, "AAAA", ["::1", "::2"], extra={"ttl": 400}
        )
        self.assertEqual(records[0].id, "AAAA:alice")
        self.assertEqual(records[0].name, "alice")
        self.assertEqual(records[0].type, RecordType.AAAA)
        self.assertEqual(records[0].data, "::1")
        self.assertEqual(records[1].id, "AAAA:alice")
        self.assertEqual(records[1].name, "alice")
        self.assertEqual(records[1].type, RecordType.AAAA)
        self.assertEqual(records[1].data, "::2")

    def test_update_multivalue_record(self):
        record = self.driver.get_record(self.test_zone.id, "MX:lists")
        updated = self.driver.update_record(
            record, None, None, "mail1", {"ttl": 400, "priority": 10}
        )
        self.assertEqual(updated.extra["priority"], "10")
        self.assertEqual(updated.data, "mail1")
        self.assertTrue("_other_records" in record.extra)
        other_record = record.extra["_other_records"][0]
        self.assertEqual(other_record["extra"]["priority"], "20")

    def test_ex_update_gandi_zone_name(self):
        updated = self.driver.ex_update_gandi_zone_name("111111", "Foo")
        self.assertTrue(updated)

    def test_ex_delete_gandi_zone(self):
        deleted = self.driver.ex_delete_gandi_zone("111111")
        self.assertTrue(deleted)


class GandiLiveMockHttp(BaseGandiLiveMockHttp):
    fixtures = DNSFileFixtures("gandi_live")

    def _json_api_v5_domains_get(self, method, url, body, headers):
        body = self.fixtures.load("list_zones.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_get(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_badexample_com_get(self, method, url, body, headers):
        body = self.fixtures.load("get_bad_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_nosuchzone_com_get(self, method, url, body, headers):
        body = self.fixtures.load("get_nonexistent_zone.json")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _json_api_v5_zones_post(self, method, url, body, headers):
        input = json.loads(body)
        if "badexample" in input["name"]:
            body = self.fixtures.load("create_bad_zone.json")
            return (httplib.CONFLICT, body, {}, httplib.responses[httplib.CONFLICT])
        else:
            body = self.fixtures.load("create_zone.json")
            return (
                httplib.OK,
                body,
                {"Location": "/zones/54321"},
                httplib.responses[httplib.OK],
            )

    def _json_api_v5_zones_111111_patch(self, method, url, body, headers):
        body = self.fixtures.load("update_gandi_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_zones_111111_delete(self, method, url, body, headers):
        return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_org_patch(self, method, url, body, headers):
        body = self.fixtures.load("create_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_get(self, method, url, body, headers):
        body = self.fixtures.load("list_records.json")
        resp_headers = {}
        if headers is not None and "Accept" in headers and headers["Accept"] == "text/plain":
            body = self.fixtures.load("list_records_bind.txt")
            resp_headers["Content-Type"] = "text/plain"
        return (httplib.OK, body, resp_headers, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_bob_A_get(self, method, url, body, headers):
        body = self.fixtures.load("get_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_none_A_get(self, method, url, body, headers):
        body = self.fixtures.load("get_nonexistent_record.json")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _json_api_v5_domains_example_com_records_lists_MX_get(self, method, url, body, headers):
        body = self.fixtures.load("get_mx_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_lists_MX_put(self, method, url, body, headers):
        body = self.fixtures.load("update_mx_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_post(self, method, url, body, headers):
        body = self.fixtures.load("create_record.json")
        return (
            httplib.OK,
            body,
            {"Location": "/zones/12345/records/alice/AAAA"},
            httplib.responses[httplib.OK],
        )

    def _json_api_v5_domains_badexample_com_records_post(self, method, url, body, headers):
        body = self.fixtures.load("create_existing_record.json")
        return (httplib.CONFLICT, body, {}, httplib.responses[httplib.CONFLICT])

    def _json_api_v5_domains_badexample_com_records_get(self, method, url, body, headers):
        return (
            httplib.INTERNAL_SERVER_ERROR,
            body,
            {},
            httplib.responses[httplib.INTERNAL_SERVER_ERROR],
        )

    def _json_api_v5_domains_badexample_com_records_jane_A_put(self, method, url, body, headers):
        body = self.fixtures.load("update_bad_record.json")
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.BAD_REQUEST])

    def _json_api_v5_domains_example_com_records_bob_A_put(self, method, url, body, headers):
        body = self.fixtures.load("update_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _json_api_v5_domains_example_com_records_bob_A_delete(self, method, url, body, headers):
        return (httplib.NO_CONTENT, "", {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest
from unittest.mock import Mock, patch

from libcloud.common.base import LibcloudConnection
from libcloud.common.openstack import OpenStackBaseConnection


class OpenStackBaseConnectionTest(unittest.TestCase):
    def setUp(self):
        self.timeout = 10
        OpenStackBaseConnection.conn_class = Mock()
        self.connection = OpenStackBaseConnection(
            "foo", "bar", timeout=self.timeout, ex_force_auth_url="https://127.0.0.1"
        )
        self.connection.driver = Mock()
        self.connection.driver.name = "OpenStackDriver"

    def tearDown(self):
        OpenStackBaseConnection.conn_class = LibcloudConnection

    def test_base_connection_timeout(self):
        self.connection.connect()
        self.assertEqual(self.connection.timeout, self.timeout)
        self.connection.conn_class.assert_called_with(
            host="127.0.0.1", secure=1, port=443, timeout=10
        )

    def test_set_microversion(self):
        self.connection.service_type = "compute"
        self.connection._ex_force_microversion = "2.67"
        headers = self.connection.add_default_headers({})
        self.assertEqual(headers["OpenStack-API-Version"], "compute 2.67")

        self.connection.service_type = "compute"
        self.connection._ex_force_microversion = "volume 2.67"
        headers = self.connection.add_default_headers({})
        self.assertNotIn("OpenStack-API-Version", headers)

        self.connection.service_type = "volume"
        self.connection._ex_force_microversion = "volume 2.67"
        headers = self.connection.add_default_headers({})
        self.assertEqual(headers["OpenStack-API-Version"], "volume 2.67")

    @patch("libcloud.common.base.ConnectionUserAndKey.request")
    def test_request(self, mock_request):
        OpenStackBaseConnection.conn_class._raw_data = ""
        OpenStackBaseConnection.default_content_type = "application/json"
        expected_response = Mock()
        mock_request.return_value = expected_response
        response = self.connection.request(
            "/path", data="somedata", headers={"h1": "v1"}, method="POST"
        )
        self.assertEqual(response, expected_response)
        mock_request.assert_called_with(
            action="/path",
            params={},
            data="somedata",
            method="POST",
            headers={"h1": "v1", "Content-Type": "application/json"},
            raw=False,
        )

    @patch("libcloud.test.common.test_openstack.OpenStackBaseConnection.connect", Mock())
    def test_connection_is_reused_when_details_dont_change(self):
        url = "https://example.com"

        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 1)

        for index in range(0, 10):
            self.connection._set_up_connection_info(url=url)
            self.assertEqual(self.connection.connect.call_count, 1)

    @patch("libcloud.test.common.test_openstack.OpenStackBaseConnection.connect", Mock())
    def test_connection_is_not_reused_when_details_change(self):
        url = "https://example.com"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 1)

        url = "https://example.com"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 1)

        url = "https://example.com:80"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 2)

        url = "http://example.com:80"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 3)

        url = "http://exxample.com:80"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 4)

        url = "http://exxample.com:81"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 5)

        url = "http://exxample.com:81"
        self.connection._set_up_connection_info(url=url)
        self.assertEqual(self.connection.connect.call_count, 5)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_ROUTE53
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.route53 import Route53DNSDriver


class Route53Tests(unittest.TestCase):
    def setUp(self):
        Route53DNSDriver.connectionCls.conn_class = Route53MockHttp
        Route53MockHttp.type = None
        self.driver = Route53DNSDriver(*DNS_PARAMS_ROUTE53)

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 10)
        self.assertTrue(RecordType.A in record_types)

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 5)

        zone = zones[0]
        self.assertEqual(zone.id, "47234")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "t.com")

    def test_list_records(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 10)

        record = records[1]
        self.assertEqual(record.name, "www")
        self.assertEqual(record.id, "A:www")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "208.111.35.173")
        self.assertEqual(record.extra["ttl"], 86400)

        record = records[3]
        self.assertEqual(record.type, RecordType.MX)
        self.assertEqual(record.data, "ASPMX.L.GOOGLE.COM.")
        self.assertEqual(record.extra["priority"], 1)

        record = records[4]
        self.assertEqual(record.type, RecordType.MX)
        self.assertEqual(record.data, "ALT1.ASPMX.L.GOOGLE.COM.")
        self.assertEqual(record.extra["priority"], 5)

        record = records[8]
        self.assertEqual(record.type, RecordType.SRV)
        self.assertEqual(record.data, "xmpp-server.example.com.")
        self.assertEqual(record.extra["priority"], 1)
        self.assertEqual(record.extra["weight"], 10)
        self.assertEqual(record.extra["port"], 5269)

    def test_get_zone(self):
        zone = self.driver.get_zone(zone_id="47234")
        self.assertEqual(zone.id, "47234")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "t.com")

    def test_get_record(self):
        record = self.driver.get_record(zone_id="47234", record_id="CNAME:wibble")
        self.assertEqual(record.name, "wibble")
        self.assertEqual(record.type, RecordType.CNAME)
        self.assertEqual(record.data, "t.com")

    def test_list_records_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        Route53MockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.list_records(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_get_zone_does_not_exist(self):
        Route53MockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_zone(zone_id="47234")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "47234")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_zone_does_not_exist(self):
        Route53MockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="4444", record_id="28536")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        Route53MockHttp.type = "RECORD_DOES_NOT_EXIST"

        rid = "CNAME:doesnotexist.t.com"
        try:
            self.driver.get_record(zone_id="47234", record_id=rid)
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_zone(self):
        zone = self.driver.create_zone(domain="t.com", type="master", ttl=None, extra=None)
        self.assertEqual(zone.id, "47234")
        self.assertEqual(zone.domain, "t.com")

    def test_create_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name="www", zone=zone, type=RecordType.A, data="127.0.0.1", extra={"ttl": 0}
        )

        self.assertEqual(record.id, "A:www")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")

    def test_create_record_zone_name(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name="", zone=zone, type=RecordType.A, data="127.0.0.1", extra={"ttl": 0}
        )

        self.assertEqual(record.id, "A:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")

    def test_create_TXT_record(self):
        """
        Check that TXT records are created in quotes
        """
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(name="", zone=zone, type=RecordType.TXT, data="test")
        self.assertEqual(record.id, "TXT:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.TXT)
        self.assertEqual(record.data, '"test"')

    def test_create_TXT_record_quoted(self):
        """
        Check that TXT values already quoted are not changed
        """
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(name="", zone=zone, type=RecordType.TXT, data='"test"')
        self.assertEqual(record.id, "TXT:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.TXT)
        self.assertEqual(record.data, '"test"')

    def test_create_SPF_record(self):
        """
        Check that SPF records are created in quotes
        """
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(name="", zone=zone, type=RecordType.SPF, data="test")
        self.assertEqual(record.id, "SPF:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.SPF)
        self.assertEqual(record.data, '"test"')

    def test_create_SPF_record_quoted(self):
        """
        Check that SPF values already quoted are not changed
        """
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(name="", zone=zone, type=RecordType.SPF, data='"test"')
        self.assertEqual(record.id, "SPF:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.SPF)
        self.assertEqual(record.data, '"test"')

    def test_create_TXT_record_escaped(self):
        """
        Check that TXT record with quotes inside are escaped correctly
        """
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name="", zone=zone, type=RecordType.TXT, data='test "with"'
        )
        self.assertEqual(record.id, "TXT:")
        self.assertEqual(record.name, "")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.TXT)
        self.assertEqual(record.data, '"test "with""')

    def test_create_multi_value_record(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.ex_create_multi_value_record(
            name="balancer",
            zone=zone,
            type=RecordType.A,
            data="127.0.0.1\n127.0.0.2",
            extra={"ttl": 0},
        )
        self.assertEqual(len(records), 2)
        self.assertEqual(records[0].id, "A:balancer")
        self.assertEqual(records[1].id, "A:balancer")
        self.assertEqual(records[0].name, "balancer")
        self.assertEqual(records[1].name, "balancer")
        self.assertEqual(records[0].zone, zone)
        self.assertEqual(records[1].zone, zone)
        self.assertEqual(records[0].type, RecordType.A)
        self.assertEqual(records[1].type, RecordType.A)
        self.assertEqual(records[0].data, "127.0.0.1")
        self.assertEqual(records[1].data, "127.0.0.2")

    def test_update_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[1]

        params = {
            "record": record,
            "name": "www",
            "type": RecordType.A,
            "data": "::1",
            "extra": {"ttle": 0},
        }
        updated_record = self.driver.update_record(**params)

        self.assertEqual(record.data, "208.111.35.173")

        self.assertEqual(updated_record.id, "A:www")
        self.assertEqual(updated_record.name, "www")
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.A)
        self.assertEqual(updated_record.data, "::1")

    def test_delete_zone(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        Route53MockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        Route53MockHttp.type = "RECORD_DOES_NOT_EXIST"
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail("Exception was not thrown")


class Route53MockHttp(MockHttp):
    fixtures = DNSFileFixtures("route53")

    def _2012_02_29_hostedzone_47234(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2012_02_29_hostedzone(self, method, url, body, headers):
        # print method, url, body, headers
        if method == "POST":
            body = self.fixtures.load("create_zone.xml")
            return (httplib.CREATED, body, {}, httplib.responses[httplib.OK])
        body = self.fixtures.load("list_zones.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2012_02_29_hostedzone_47234_rrset(self, method, url, body, headers):
        body = self.fixtures.load("list_records.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2012_02_29_hostedzone_47234_rrset_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _2012_02_29_hostedzone_4444_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _2012_02_29_hostedzone_47234_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.xml")
        return (httplib.NOT_FOUND, body, {}, httplib.responses[httplib.NOT_FOUND])

    def _2012_02_29_hostedzone_47234_rrset_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        if method == "POST":
            body = self.fixtures.load("invalid_change_batch.xml")
            return (
                httplib.BAD_REQUEST,
                body,
                {},
                httplib.responses[httplib.BAD_REQUEST],
            )
        body = self.fixtures.load("record_does_not_exist.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2012_02_29_hostedzone_47234_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.providers import get_driver as _get_provider_driver
from libcloud.common.providers import set_driver as _set_provider_driver
from libcloud.loadbalancer.types import OLD_CONSTANT_TO_NEW_MAPPING, Provider

__all__ = [
    "Provider",
    "DRIVERS",
    "get_driver",
]

DRIVERS = {
    Provider.RACKSPACE: (
        "libcloud.loadbalancer.drivers.rackspace",
        "RackspaceLBDriver",
    ),
    Provider.NINEFOLD: ("libcloud.loadbalancer.drivers.ninefold", "NinefoldLBDriver"),
    Provider.BRIGHTBOX: (
        "libcloud.loadbalancer.drivers.brightbox",
        "BrightboxLBDriver",
    ),
    Provider.ELB: ("libcloud.loadbalancer.drivers.elb", "ElasticLBDriver"),
    Provider.ALB: ("libcloud.loadbalancer.drivers.alb", "ApplicationLBDriver"),
    Provider.CLOUDSTACK: (
        "libcloud.loadbalancer.drivers.cloudstack",
        "CloudStackLBDriver",
    ),
    Provider.GCE: ("libcloud.loadbalancer.drivers.gce", "GCELBDriver"),
    Provider.DIMENSIONDATA: (
        "libcloud.loadbalancer.drivers.dimensiondata",
        "DimensionDataLBDriver",
    ),
    Provider.ALIYUN_SLB: ("libcloud.loadbalancer.drivers.slb", "SLBDriver"),
    Provider.NTTCIS: ("libcloud.loadbalancer.drivers.nttcis", "NttCisLBDriver"),
}


def get_driver(provider):
    deprecated_constants = OLD_CONSTANT_TO_NEW_MAPPING
    return _get_provider_driver(
        drivers=DRIVERS, provider=provider, deprecated_constants=deprecated_constants
    )


def set_driver(provider, module, klass):
    return _set_provider_driver(drivers=DRIVERS, provider=provider, module=module, klass=klass)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.utils.py3 import b, httplib, base64_encode_string
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.compute.types import InvalidCredsError

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore


class BrightboxResponse(JsonResponse):
    def success(self):
        return httplib.OK <= self.status < httplib.BAD_REQUEST

    def parse_body(self):
        if self.headers["content-type"].split(";")[0] == "application/json":
            return super().parse_body()
        else:
            return self.body

    def parse_error(self):
        response = super().parse_body()

        if "error" in response:
            if response["error"] in ["invalid_client", "unauthorized_client"]:
                raise InvalidCredsError(response["error"])

            return response["error"]
        elif "error_name" in response:
            return "{}: {}".format(response["error_name"], response["errors"][0])

        return self.body


class BrightboxConnection(ConnectionUserAndKey):
    """
    Connection class for the Brightbox driver
    """

    host = "api.gb1.brightbox.com"
    responseCls = BrightboxResponse

    def _fetch_oauth_token(self):
        body = json.dumps({"client_id": self.user_id, "grant_type": "none"})

        authorization = (
            "Basic " + str(base64_encode_string(b("{}:{}".format(self.user_id, self.key)))).rstrip()
        )

        self.connect()

        headers = {
            "Host": self.host,
            "User-Agent": self._user_agent(),
            "Authorization": authorization,
            "Content-Type": "application/json",
            "Content-Length": str(len(body)),
        }

        # pylint: disable=assignment-from-no-return
        response = self.connection.request(method="POST", url="/token", body=body, headers=headers)

        if response.status == httplib.OK:
            return json.loads(response.read())["access_token"]
        else:
            responseCls = BrightboxResponse(response=response.getresponse(), connection=self)
            message = responseCls.parse_error()
            raise InvalidCredsError(message)

    def add_default_headers(self, headers):
        try:
            headers["Authorization"] = "OAuth " + self.token
        except AttributeError:
            self.token = self._fetch_oauth_token()

            headers["Authorization"] = "OAuth " + self.token

        return headers

    def encode_data(self, data):
        return json.dumps(data)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Dict, List
from xml.etree import ElementTree as ET  # noqa

from libcloud.common.base import XmlResponse, ConnectionUserAndKey

# API HOST to connect
API_HOST = "durabledns.com"


def _schema_builder(urn_nid, method, attributes):
    """
    Return a xml schema used to do an API request.

    :param urn_nid: API urn namespace id.
    :type urn_nid: type: ``str``

    :param method: API method.
    :type method: type: ``str``

    :param attributes: List of attributes to include.
    :type attributes: ``list`` of ``str``

    rtype: :class:`Element`
    """
    soap = ET.Element("soap:Body", {"xmlns:m": "https://durabledns.com/services/dns/%s" % method})
    urn = ET.SubElement(soap, "urn:{}:{}".format(urn_nid, method))
    # Attributes specification
    for attribute in attributes:
        ET.SubElement(urn, "urn:{}:{}".format(urn_nid, attribute))
    return soap


SCHEMA_BUILDER_MAP = {
    "list_zones": {
        "urn_nid": "listZoneswsdl",
        "method": "listZones",
        "attributes": ["apiuser", "apikey"],
    },
    "list_records": {
        "urn_nid": "listRecordswsdl",
        "method": "listRecords",
        "attributes": ["apiuser", "apikey", "zonename"],
    },
    "get_zone": {
        "urn_nid": "getZonewsdl",
        "method": "getZone",
        "attributes": ["apiuser", "apikey", "zonename"],
    },
    "get_record": {
        "urn_nid": "getRecordwsdl",
        "method": "getRecord",
        "attributes": ["apiuser", "apikey", "zonename", "recordid"],
    },
    "create_zone": {
        "urn_nid": "createZonewsdl",
        "method": "createZone",
        "attributes": [
            "apiuser",
            "apikey",
            "zonename",
            "ns",
            "mbox",
            "refresh",
            "retry",
            "expire",
            "minimum",
            "ttl",
            "xfer",
            "update_acl",
        ],
    },
    "create_record": {
        "urn_nid": "createRecordwsdl",
        "method": "createRecord",
        "attributes": [
            "apiuser",
            "apikey",
            "zonename",
            "name",
            "type",
            "data",
            "aux",
            "ttl",
            "ddns_enabled",
        ],
    },
    "update_zone": {
        "urn_nid": "updateZonewsdl",
        "method": "updateZone",
        "attributes": [
            "apiuser",
            "apikey",
            "zonename",
            "ns",
            "mbox",
            "refresh",
            "retry",
            "expire",
            "minimum",
            "ttl",
            "xfer",
            "update_acl",
        ],
    },
    "update_record": {
        "urn_nid": "updateRecordwsdl",
        "method": "updateRecord",
        "attributes": [
            "apiuser",
            "apikey",
            "zonename",
            "id",
            "name",
            "aux",
            "data",
            "ttl",
            "ddns_enabled",
        ],
    },
    "delete_zone": {
        "urn_nid": "deleteZonewsdl",
        "method": "deleteZone",
        "attributes": ["apiuser", "apikey", "zonename"],
    },
    "delete_record": {
        "urn_nid": "deleteRecordwsdl",
        "method": "deleteRecord",
        "attributes": ["apiuser", "apikey", "zonename", "id"],
    },
}


class DurableDNSException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "DurableDNSException {} {}".format(self.code, self.message)


class DurableResponse(XmlResponse):
    errors = []  # type: List[Dict]
    objects = []  # type: List[Dict]

    def __init__(self, response, connection):
        super().__init__(response=response, connection=connection)

        self.objects, self.errors = self.parse_body_and_error()
        if self.errors:
            raise self._make_excp(self.errors[0])

    def parse_body_and_error(self):
        """
        Used to parse body from httplib.HttpResponse object.
        """
        objects = []
        errors = []
        error_dict = {}
        extra = {}
        zone_dict = {}
        record_dict = {}
        xml_obj = self.parse_body()

        # pylint: disable=no-member
        envelop_body = list(xml_obj)[0]
        method_resp = list(envelop_body)[0]
        # parse the xml_obj
        # handle errors
        if "Fault" in method_resp.tag:
            fault = [fault for fault in list(method_resp) if fault.tag == "faultstring"][0]
            error_dict["ERRORMESSAGE"] = fault.text.strip()
            error_dict["ERRORCODE"] = self.status
            errors.append(error_dict)

        # parsing response from listZonesResponse
        if "listZonesResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            for element in answer:
                zone_dict["id"] = list(element)[0].text
                objects.append(zone_dict)
                # reset the zone_dict
                zone_dict = {}
        # parse response from listRecordsResponse
        if "listRecordsResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            for element in answer:
                for child in list(element):
                    if child.tag == "id":
                        record_dict["id"] = child.text.strip()
                objects.append(record_dict)
                # reset the record_dict for later usage
                record_dict = {}
        # parse response from getZoneResponse
        if "getZoneResponse" in method_resp.tag:
            for child in list(method_resp):
                if child.tag == "origin":
                    zone_dict["id"] = child.text.strip()
                    zone_dict["domain"] = child.text.strip()
                elif child.tag == "ttl":
                    zone_dict["ttl"] = int(child.text.strip())
                elif child.tag == "retry":
                    extra["retry"] = int(child.text.strip())
                elif child.tag == "expire":
                    extra["expire"] = int(child.text.strip())
                elif child.tag == "minimum":
                    extra["minimum"] = int(child.text.strip())
                else:
                    if child.text:
                        extra[child.tag] = child.text.strip()
                    else:
                        extra[child.tag] = ""
                    zone_dict["extra"] = extra
            objects.append(zone_dict)
        # parse response from getRecordResponse
        if "getRecordResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            for child in list(method_resp):
                if child.tag == "id" and child.text:
                    record_dict["id"] = child.text.strip()
                elif child.tag == "name" and child.text:
                    record_dict["name"] = child.text.strip()
                elif child.tag == "type" and child.text:
                    record_dict["type"] = child.text.strip()
                elif child.tag == "data" and child.text:
                    record_dict["data"] = child.text.strip()
                elif child.tag == "aux" and child.text:
                    record_dict["aux"] = child.text.strip()
                elif child.tag == "ttl" and child.text:
                    record_dict["ttl"] = child.text.strip()
            if not record_dict:
                error_dict["ERRORMESSAGE"] = "Record does not exist"
                error_dict["ERRORCODE"] = 404
                errors.append(error_dict)
            objects.append(record_dict)
            record_dict = {}
        if "createZoneResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            if answer.tag == "return" and answer.text:
                record_dict["id"] = answer.text.strip()
            objects.append(record_dict)
        # catch Record does not exists error when deleting record
        if "deleteRecordResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            if "Record does not exists" in answer.text.strip():
                errors.append({"ERRORMESSAGE": answer.text.strip(), "ERRORCODE": self.status})
        # parse response in createRecordResponse
        if "createRecordResponse" in method_resp.tag:
            answer = list(method_resp)[0]
            record_dict["id"] = answer.text.strip()
            objects.append(record_dict)
            record_dict = {}

        return (objects, errors)

    def parse_body(self):
        # A problem arise in the api response because there are undeclared
        # xml namespaces. In order to fix that at the moment, we use the
        # _fix_response method to clean up since we won't always have lxml
        # library.
        self._fix_response()
        body = super().parse_body()
        return body

    def success(self):
        """
        Used to determine if the request was successful.
        """
        return len(self.errors) == 0

    def _make_excp(self, error):
        return DurableDNSException(error["ERRORCODE"], error["ERRORMESSAGE"])

    def _fix_response(self):
        items = re.findall('<ns1:.+ xmlns:ns1="">', self.body, flags=0)
        for item in items:
            parts = item.split(" ")
            prefix = parts[0].replace("<", "").split(":")[1]
            new_item = "<" + prefix + ">"
            close_tag = "</" + parts[0].replace("<", "") + ">"
            new_close_tag = "</" + prefix + ">"
            self.body = self.body.replace(item, new_item)
            self.body = self.body.replace(close_tag, new_close_tag)


class DurableConnection(ConnectionUserAndKey):
    host = API_HOST
    responseCls = DurableResponse

    def add_default_params(self, params):
        params["user_id"] = self.user_id
        params["key"] = self.key
        return params

    def add_default_headers(self, headers):
        headers["Content-Type"] = "text/xml"
        headers["Content-Encoding"] = "gzip; charset=ISO-8859-1"
        return headers
